'''Functiona graph analysis tools

'''
from pathlib import Path
import json
import hashlib
import logging
import re
import abc
import types
import typing as T

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

import toolz.curried as _
import networkx as nx

from .. import common as __

class _graph_null(__._null):
    _instance = None

    def __repr__(self):
        return 'GraphNull'

    def __gt__(self, function):
        return function(self)

GraphNull = _graph_null()
    
GraphFunction = T.Callable[['...'], nx.Graph]

def node_type(graph, node_id):
    return graph.nodes[node_id]['type']

_traversal_functions = set()
def traversal_function(func):
    _traversal_functions.add(func.__name__)
    return func

def is_traversal_function(function):
    return function.__name__ in _traversal_functions
        
_introspect_functions = set()
def introspect_function(func):
    _introspect_functions.add(func.__name__)
    return func

def is_introspect_function(function):
    return function.__name__ in _introspect_functions

_edge_functions = set()
def edge_function(func):
    _edge_functions.add(func.__name__)
    return func
        
Selector = T.Callable[[nx.Graph, T.Any, dict], bool]

@_.curry
@traversal_function
def neighbor_selector(selectors: T.Iterable[Selector],
                      graph: nx.Graph, node_id: T.Any,
                      **attrs) -> T.Iterable[T.Tuple[T.Any, dict]]:
    if attrs:
        selectors = _.concatv(
            selectors, [lambda graph, node_id, edge_data: all(
                v(graph.nodes[node_id][k]) for k, v in attrs.items()
            )],
        )
    return _.pipe(
        graph[node_id].items(),
        __.vfilter(lambda node_id, edge_data: all(
            s(graph, node_id, edge_data) for s in selectors
        )),
    )

@_.curry
@traversal_function
def neighbors_of_type(type, graph, node_id, *,
                      edge_selector=lambda n, e: True):
    return _.pipe(
        graph[node_id].items(),
        __.vfilter(lambda nid, edata: node_type(graph, nid) == type),
        __.vfilter(edge_selector),
        _.map(_.first),
    )

@_.curry
@introspect_function
def get(key: str, graph: nx.Graph, node_id: T.Any):
    return _.pipe(
        graph.nodes[node_id],
        __.jmes(key),
    )

@_.curry
@edge_function
def e_get(key: str, graph: nx.Graph, v0: T.Any, v1: T.Any):
    return _.pipe(
        graph.edges,
    )

@_.curry
@introspect_function
def items(graph: nx.Graph, node_id: T.Any):
    return graph.nodes[node_id].items()

FunctionSpace = T.Dict[str, T.Callable]

def to_function_space(f_space: T.Union[type, FunctionSpace,
                                       T.Iterable[T.Callable]]):
    if __.is_dict(f_space):
        return f_space
    if __.is_seq(f_space):
        return {
            f.__name__: f for f in f_space
        }
    return _.pipe(
        f_space.__dict__.items(),
        __.vfilter(lambda k, v: not (k.startswith('__') and k.endswith('__'))),
        dict,
    )

def graph_pipe(*function_spaces: FunctionSpace):
    @_.curry
    def pipe(graph, selection):
        return GraphPipe(
            graph, selection,
            *_.pipe(
                _.concatv(
                    [{
                        'neighbors_of_type': neighbors_of_type,
                        'neighbor_selector': neighbor_selector,
                        'get': get,
                        'items': items,
                    }],
                    function_spaces,
                ),
                _.map(to_function_space),
            )
        )
    return pipe

class GraphPipe:
    def __init__(self, graph, selection, *function_spaces):
        self.graph = graph
        self.selection = selection
        self.function_space = _.merge(*function_spaces)

    def __or__(self, function):
        if __.is_str(function):
            # function = globals()[function]
            log.debug(f'function space: {self.function_space}')
            function = eval(function, self.function_space)

        if is_traversal_function(function):
            nids, _neighbors = zip(*self.selection)

            return _.pipe(
                # self.selection,
                # _.mapcat(_.second),
                nids,
                # _.concat,
                set,
                # _.do(print),
                _.map(function(self.graph)),
                _.mapcat(tuple),
                set,
                lambda node_ids: self.graph(*node_ids),
            )
        elif is_introspect_function(function):
            node_ids, _nb = zip(*self.selection)
            return _.pipe(
                node_ids,
                set,
                _.map(function(self.graph)),
                ValuePipe,
            )
        else:
            node_ids, _nb = zip(*self.selection)
            return ValuePipe(function(node_ids))
            
    def __gt__(self, function):
        node_ids, _nb = zip(*self.selection)
        return function(node_ids)

class ValuePipe:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return repr(self.value)

    def __or__(self, other):
        if callable(other):
            return ValuePipe(other(self.value))
        return self.value | other

    def __gt__(self, function):
        return function(self.value)

class Graph(nx.DiGraph):
    def __init__(self, data, graph_pipe, **attr):
        self.graph_pipe = graph_pipe
        super().__init__(data, **attr)

    def __call__(self, *node_ids):
        if len(node_ids) == 1 and __.is_seq(node_ids[0]):
            node_ids = tuple(node_ids[0])

        bad_ids = _.pipe(
            node_ids,
            _.filter(lambda i: i not in self),
            tuple,
        )
        if bad_ids:
            bad_str = _.pipe(bad_ids, _.map(str), ', '.join)
            log.error(f'Bad node ids: {bad_str}')

        # log.info(type(node_ids[0]))
        node_ids = set(node_ids) - set(bad_ids)
        # log.info(node_ids)
        if not node_ids:
            log.error('No ids left')
            return GraphNull
            
        return _.pipe(
            node_ids,
            _.map(lambda n: (n, self[n])),
            self.graph_pipe(self),
        )
    
    def of_type(self, type):
        return _.pipe(
            self,
            _.filter(lambda n: self.nodes[n]['type'] == type),
            _.map(lambda n: (n, self[n])),
            self.graph_pipe(self),
        )

    # @_.curry
    # @classmethod
    # def from_graph_function(cls, function: GraphFunction, *a, **kw):
    #     return cls(function(*a, **kw))




import toolz.curried as _
import markdown as _markdown

from . import (
    meta_yaml, card, table, yaml_data,
)

@_.curry
def markdown(content: str, **kwargs):
    
    class HtmlWithMeta(str):
        meta = None

    md = _markdown.Markdown(
        extensions=_.pipe(
            _.concatv(
                ['larc.markdown.meta_yaml',
                 'larc.markdown.yaml_data',
                 'larc.markdown.card',
                 'larc.markdown.table'],
                ['extra', 'codehilite', 'toc', 'admonition'],
                kwargs.get('extensions', []),
            ),
            set,
            tuple,
        ),
            
        extension_configs=_.merge(
            {
                'extra': {},
                'admonition': {
                },
                'codehilite': {
                    'noclasses': True,
                    'guess_lang': False,
                },
            },
            kwargs.get('extension_configs', {}),
        ),
    )

    output = HtmlWithMeta(md.convert(content))
    output.meta = md.meta or {}
    return output



import re
from pathlib import Path
import calendar

import jinja2
import toolz.curried as _

from .. import common as __
from .. import logging
from .. import images
from .. import markdown

log = logging.new_log(__name__)

@_.curry
def find_files(start, filename):
    start_path = Path(start).resolve()

    log.info(f'Finding image {filename}')
    if Path(start_path, filename).exists():
        path = Path(start_path, filename)
        log.info(f'   found: {path}')
        return [path]

    file_re = re.compile(
        filename.lower().replace('.', '\\.').replace('*', '.*')
    )
    log.info(f'   searching with re: {file_re.pattern}')

    paths = _.pipe(
        __.walk(start_path),
        filter(lambda p: file_re.search(p.name.lower())),
        _.partial(sorted, key=lambda p: p.name),
        tuple,
    )
    if not paths:
        log.error(
            f'Could not find any filenames meeting the search string'
            f' "{filename}" looking in directory: {start_path}'
        )
    return paths

def pluralize(number, singular='', plural='s'):
    if number == 1:
        return singular
    return plural

def md_to_html(text):
    return markdown.markdown(text)

def long_date(text):
    import datetime
    import dateutil.parser
    try:
        if isinstance(text, datetime.datetime):
            dt = text
        else:
            dt = dateutil.parser.parse(text)
        return dt.strftime('%B %d, %Y')
    except ValueError as error:
        log.error(
            'Error in long_date:\n'
            f'error: {error}\n'
            f'for text: {repr(text)}\n'
            'exception:', exc_info=True
        )

def year(text):
    return _.pipe(
        text,
        __.maybe_dt,
        __.deref('year'),
    )

def month(text):
    return _.pipe(
        text,
        __.maybe_dt,
        __.deref('month'),
        lambda m: calendar.month_name[m],
    )

def add_filters(env: jinja2.Environment):
    env.filters['pluralize'] = pluralize
    env.filters['md_to_html'] = md_to_html
    env.filters['long_date'] = long_date
    env.filters['year'] = year
    env.filters['month'] = month
    env.filters['b64encode'] = __.b64encode_str
    return env

@_.curry
def make_figure(im_type, filename, caption, images, *, style: dict = None):
    style = _.pipe(
        _.pipe(
            style.items(),
            __.vmap(lambda k, v: f"{k}: {v}"),
            ';'.join,
        ),
        'style="{}"'.format,
    ) if style else ''
    figure = {'type': im_type, 'filename': Path(filename).name,
              'search': filename, 'caption': caption,
              'images': images, 'style': style}
    return get_env().get_template('figure.html.j2').render(figure=figure)

@_.curry
def figure_function(im_type, bytes_function, caption, filename, start='.',
                    *, style: dict = None):
    return _.pipe(
        filename,
        find_files(start),
        map(bytes_function),
        map(__.b64encode_str),
        tuple,
        make_figure(im_type, filename, caption, style=style),
    )

png = figure_function('png', images.common.png_bytes)
jpeg = figure_function('jpeg', images.common.jpeg_bytes)

def table(data):
    return get_env().get_template('table.html.j2').render(data=data)

def add_functions(env: jinja2.Environment):
    env.globals['png'] = png
    env.globals['jpeg'] = jpeg
    env.globals['table'] = table
    return env

get_env = _.compose_left(
    jinja2.Environment,
    add_filters,
    add_functions,
)



'''Tools for dealing with text data
'''
from pathlib import Path
import sys

import jinja2
import click
from .. import common as __
import toolz.curried as _

from ..common import difflines, intlines

from .. import markdown

@click.command(
    help=('Given PATHA and PATHB of text, print difference (A-B)'),
)
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def diff_lines(patha, pathb):
    _.pipe(
        difflines(
            Path(patha).expanduser().read_text(),
            Path(pathb).expanduser().read_text(),
        ),
        '\n'.join,
        print,
    )

@click.command(
    help=('Given PATHA and PATHB of text, print intersection (A & B)'),
)
@click.argument(
    'patha',
    type=click.Path(exists=True),
)
@click.argument(
    'pathb',
    type=click.Path(exists=True),
)
def int_lines(patha, pathb):
    _.pipe(
        intlines(
            Path(patha).expanduser().read_text(),
            Path(pathb).expanduser().read_text(),
        ),
        '\n'.join,
        print,
    )


@click.command(
    help=(
        'Transform text via parsing inline YAML'
        ' metadata and rendering as template through Jinja2'
    ),
)
@click.argument(
    'paths', nargs=-1,
)
def render_templates(paths):
    contents = _.pipe(
        paths,
        _.map(lambda p: Path(p).expanduser()),
        _.map(lambda p: p.read_text()),
        tuple,
    )

    if not contents:
        contents = [sys.stdin.read()]
    
    _.pipe(
        contents,
        _.map(markdown.meta_yaml.meta_and_content),
        __.vmap(lambda d, c: jinja2.Environment().from_string(c).render(**d)),
        _.map(_.do(print)),
        tuple,
    )



import networkx as nx
import toolz.curried as _

from .. import common as __
from . import common

class node_types:
    hostname = 'hostname'
    ip = 'ip'
    local_admin = 'local_admin'
    admin = local_admin
    group = 'group'
    user = 'user'
    session = 'session'
    domain = 'domain'
    trust = 'trust'
    finding = 'finding'
    finding_i = 'finding_i'
    vuln = finding
    vuln_i = finding_i
    scan = 'scan'
    software = 'software'
    tag = 'tag'
    label = 'tag'
    endpoint = 'endpoint'
    port = 'port'
    service = 'service'
    fingerprint = 'fingerprint'
    os = 'os'
    likely_os = os

    relations = (
        (ip, 'many', hostname),
        (hostname, 'many', ip),

        (ip, 'many', local_admin),
        (hostname, 'many', local_admin),

        (ip, 'many', session),
        (hostname, 'many', session),

        (ip, 'many', finding),
        (hostname, 'many', finding),

        (ip, 'many', fingerprint),
        (hostname, 'many', fingerprint),

        (ip, 'many', software),
        (hostname, 'many', software),

        (ip, 'one', domain),
        (domain, 'many', ip),
        (hostname, 'one', domain),
        (domain, 'many', hostname),

        (local_admin, 'one', user),

        (domain, 'many', group),
        (group, 'one', domain),
        
        (domain, 'many', user),
        (user, 'one', domain),
        
        (domain, 'many', trust),
        (trust, 'one', domain),

        (finding, 'many', tag),
        (tag, 'many', finding),
        
        (ip, 'many', endpoint),
        (hostname, 'many', endpoint),

        (endpoint, 'one', ip),
        (endpoint, 'one', hostname),
        (endpoint, 'one', port),
        (endpoint, 'many', service),
        (endpoint, 'many', finding),

        (port, 'many', endpoint),
        
        (service, 'many', fingerprint),
    )

type_selectors = {
    aname: common.neighbors_of_type(getattr(node_types, aname))
    for aname in node_types.__dict__
    if not aname.startswith('_') and aname != 'relations'
}

# class type_selectors:
#     for aname in node_types.__dict__:
#         if not aname.startswith('_') and aname != 'relations':
#             attr = getattr(node_types, aname)
#             locals()[aname] = common.neighbors_of_type(attr)
            
#     # fingerprint = common.neighbors_of_type(node_types.fingerprint)
#     # host = common.neighbors_of_type(node_types.host)
#     # hostname = common.neighbors_of_type(node_types.hostname)
#     name = hostname
#     finding = common.neighbors_of_type(node_types.finding)
#     vuln = finding
#     tag = common.neighbors_of_type(node_types.tag)
#     port = common.neighbors_of_type(node_types.port)

def graph_pipe(*function_spaces: common.FunctionSpace):
    return common.graph_pipe(*_.concatv([type_selectors], function_spaces))

def new_graph(graph, *function_spaces):
    return common.Graph(graph, graph_pipe(*function_spaces))



import io
import logging
from functools import wraps
from pathlib import Path
import typing as T

from pkg_resources import resource_filename
import toolz.curried as _
from multipledispatch import dispatch
from PIL import (
    Image,
    ImageDraw,
    ImageFont,
)

from .. import common as __
from ..logging import new_log

log = new_log(__name__)

BLUE = (87, 116, 160, 255)
BLUE2 = (100, 140, 200, 255)
GREY = (100, 100, 100, 255)

@dispatch((str, Path))
def jpeg_bytes(path: T.Union[str, Path]):
    return jpeg_bytes(Image.open(path))

@dispatch(Image.Image)          # noqa
def jpeg_bytes(image: Image.Image):
    buf = io.BytesIO()
    image.convert('RGB').save(buf, 'jpeg', optimize=True)
    return buf.getvalue()

@dispatch((str, Path))
def png_bytes(path: T.Union[str, Path]):
    return png_bytes(Image.open(path))

@dispatch(Image.Image)          # noqa
def png_bytes(image: Image.Image):
    buf = io.BytesIO()
    image.save(buf, 'png')
    return buf.getvalue()

def delta_point(x, y):
    def delta(dx=0, dy=0):
        return (x + dx, y + dy)
    return delta

def get_font(name):
    return resource_filename(__name__, f'templates/fonts/{name}')

def tt_font(path):
    def font(*a, **kw):
        return ImageFont.truetype(path, *a, **kw)
    return font

def cambria():
    fmap = _.pipe(
        [(frozenset(['italic', 'bold']), 'Cambria Bold Italic.ttf'),
         (frozenset(['bold']), 'Cambria Bold.ttf'),
         (frozenset(['italic']), 'Cambria Italic.ttf'),
         (frozenset(['math']), 'Cambria Math.ttf'),
         (frozenset(), 'Cambria.ttf')],
        _.map(lambda d: (d[0], tt_font(get_font(d[1])))),
        dict,
    )

    def font(*mods):
        return fmap[frozenset(mods)]
    return font

@_.curry
def draw_text(base_image: Image.Image, text: str, loc: T.Tuple[int, int], *,
              font_f: T.Callable = cambria, font_size: int = 20,
              mods: T.Iterable = None, size: T.Tuple[int, int] = None,
              fill: T.Union[str, T.Tuple[int, int, int]] = 'white'):
    overlay = Image.new('RGBA', base_image.size, (255, 255, 255, 0))
    ovl_draw = ImageDraw.Draw(overlay)

    font = font_f()(*(mods or []))(font_size)
    loc = tuple(loc)

    draw_method = ovl_draw.text
    if '\n' in text:
        draw_method = ovl_draw.multiline_text
        
    draw_method(loc, text, font=font, fill=fill)

    final = Image.alpha_composite(base_image, overlay)

    if size:
        sx, sy = size
        w, h = final.size
        final = final.resize(
            (int(w * sx), int(h * sy)), resample=Image.LANCZOS
        )
        
    return final



import logging
from pathlib import Path
from functools import wraps
import typing as T
from collections.abc import Mapping, Iterable

import ruamel.yaml
from ruamel.yaml.comments import CommentedMap
from toolz import pipe
from multipledispatch import dispatch

from . import common

log = logging.getLogger('yaml')
log.addHandler(logging.NullHandler())

@wraps(ruamel.yaml.dump)
def dump(*a, **kw) -> str:
    kw['Dumper'] = ruamel.yaml.RoundTripDumper
    kw['default_flow_style'] = False
    kw['width'] = 2**31
    return ruamel.yaml.dump(*a, **kw)

@wraps(ruamel.yaml.load)
def load(*a, **kw) -> T.Any:
    kw['Loader'] = ruamel.yaml.RoundTripLoader
    return ruamel.yaml.load(*a, **kw)
    
def read_yaml(path: (str, Path)):
    '''Read YAML data from path and return object
    '''
    with Path(path).expanduser().open() as rfp:
        return load(rfp)

def maybe_read_yaml(path: (str, Path)):
    try:
        return read_yaml(path)
    except Exception:
        log.exception(f'Error reading YAML path: {path}')
        return common.Null

def _write_yaml(path, data):
    with Path(path).open('w') as wfp:
        dump(data, wfp)
    return True

@dispatch((str, Path), Mapping)
def write_yaml(path, dict_data):
    '''Write data as YAML to path

    Args:
      path (str, Path): path to write to

      data (dict-like): dictionary-like object to write (will be
         recursively converted to base Python types)

    Returns: (bool) success of write operation,

    Raises: on error, will raise exception

    '''
    return _write_yaml(path, pipe(dict_data,
                                  common.no_pyrsistent,
                                  CommentedMap))

@dispatch((str, Path), Iterable)  # noqa
def write_yaml(path, iterable_data):
    '''Write data as YAML to path

    Args:
      path (str, Path): path to write to

      data (sequence): sequence object to write (will be recursively
         converted to base Python types)

    Returns: (bool) success of write operation,

    Raises: on error, will raise exception

    '''
    return _write_yaml(path, pipe(iterable_data, common.no_pyrsistent))

@dispatch((str, Path), object)  # noqa
def write_yaml(path, object_data):
    return _write_yaml(path, object_data)

